from simulation import *
import numpy as np
from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *
from simobjects.pointlight import *
from simobjects.pointcloud import *
from utils import *
import robot as rb
from scipy.linalg import null_space
from graphics.meshmaterial import *
import sys


def _control_demo_2():

    # Create simulation and add objects to the scene
    material_box = MeshMaterial(color="#242526", roughness=1, metalness=1)
    material_glass = MeshMaterial(metalness=0.9, roughness=0.05, env_map_intensity=0.9, clearcoat=1, opacity=1, \
                              reflectivity=0.2, refraction_ratio=0.985, ior=1.52, specular_intensity=0.1,
                              specular_color="white", transmission=1, side="BackSide")

    robot_a = rb.Robot.create_kuka_kr5(name="robot_a", htm=Utils.trn([0, -0.9, 0.3]), color="#df6c25")
    box_a = Box(name="box_a", width=0.3, depth=0.3, height=0.3, htm=Utils.trn([0, -0.9, 0.15]), mesh_material=material_box)

    robot_b = rb.Robot.create_kuka_kr5(name="robot_b", htm=Utils.rotz(np.pi) * Utils.trn([0, -0.9, 0.3]),
                                       color="#4d6fc4")
    box_b = Box(name="box_b", width=0.3, depth=0.3, height=0.3, htm=Utils.trn([0, 0.9, 0.15]), mesh_material=material_box)

    box_left = Box(name="box_left", width=0.3, depth=0.3, height=0.366, htm=Utils.trn([0.3, 0, 0.183]),
                   mesh_material=material_box)
    box_right = Box(name="box_right", width=0.3, depth=0.3, height=0.366, htm=Utils.trn([-0.3, 0, 0.183]),
                    mesh_material=material_box)

    plate = Box(name="plate", width=0.2, depth=0.489, height=0.02, htm=Utils.trn([-0.3, 0, 0.391]),
                mesh_material=material_glass) #height=0.05

    sim = Simulation.create_sim_factory([robot_a, robot_b, box_a, box_b, box_left, box_right, plate])

    # Use inverse kinematics to set the robot to the starting pose
    q_a = robot_a.ikm(Utils.trn([-0.3, 0.1, -0.7]) * robot_a.fkm())
    q_b = robot_b.ikm(Utils.trn([-0.3, -0.1, -0.7]) * robot_b.fkm())

    robot_a.add_ani_frame(0, q_a)
    robot_b.add_ani_frame(0, q_b)

    mth_a_b_des = np.linalg.inv(robot_a.fkm()) * robot_b.fkm()
    mth_plate = np.linalg.inv(robot_a.fkm()) * plate.htm

    # Simulation parameters
    dt = 0.005
    time_max_1 = 3  # 1
    time_max_2 = 4  # 1.5

    jmax = round((2 * time_max_1 + time_max_2) / dt)
    K = 1

    # Initializations
    hist_time = []
    hist_qdot_a = np.matrix(np.zeros((6,0)))
    hist_q_a = np.matrix(np.zeros((6,0)))
    hist_qdot_b = np.matrix(np.zeros((6,0)))
    hist_q_b = np.matrix(np.zeros((6,0)))
    hist_error_rel_pos = np.matrix(np.zeros((3,0)))
    hist_error_rel_ori = np.matrix(np.zeros((3,0)))
    hist_time.append(0)

    # Main loop
    # First part: lift plate
    mth_a_des = Utils.trn([0, 0, 0.1]) * robot_a.fkm()
    j = 0
    for i in range(round(time_max_1 / dt)):

        if j % 50 == 0 or j == jmax - 1:
            sys.stdout.write('\r')
            sys.stdout.write("[%-20s] %d%%" % ('=' * round(20 * j / (jmax - 1)), round(100 * j / (jmax - 1))))
            sys.stdout.flush()

        r, jac_r = rb.Robot.coop_task_function(robot_a, robot_b, mth_a_des, mth_a_b_des, q_a, q_b)

        r[3:6] = np.sqrt(r[3:6])
        r[9:12] = np.sqrt(r[9:12])

        u = Utils.hierarchical_solve([jac_r[0:6, :], jac_r[6:12, :]],[-K * r[0:6], -K * r[6:12]])

        q_a += u[0:6] * dt
        q_b += u[6:12] * dt

        robot_a.add_ani_frame(j * dt, q_a)
        robot_b.add_ani_frame(j * dt, q_b)
        plate.add_ani_frame(j * dt, robot_a.fkm() * mth_plate)

        hist_time.append(hist_time[-1] + dt)
        hist_q_a = np.block([hist_q_a, robot_a.q])
        hist_qdot_a = np.block([hist_qdot_a, u[0:6]])
        hist_q_b  = np.block([hist_q_b, robot_b.q])
        hist_qdot_b = np.block([hist_qdot_b, u[6:12]])
        hist_error_rel_pos = np.block([hist_error_rel_pos, r[0:3]])
        error_ori = np.matrix([(180 / (np.pi)) * acos(1 - num * num) for num in r[3:6]]).reshape((3,1))
        hist_error_rel_ori = np.block([hist_error_rel_ori, error_ori])

        j += 1

    # Second part: move to the left (negative x axis)
    mth_a_des = Utils.trn([0.6, 0, 0]) * robot_a.fkm()
    for i in range(round(time_max_2 / dt)):

        if j % 50 == 0 or j == jmax - 1:
            sys.stdout.write('\r')
            sys.stdout.write("[%-20s] %d%%" % ('=' * round(20 * j / (jmax - 1)), round(100 * j / (jmax - 1))))
            sys.stdout.flush()

        r, jac_r = rb.Robot.coop_task_function(robot_a, robot_b, mth_a_des, mth_a_b_des, q_a, q_b)

        r[3:6] = np.sqrt(r[3:6])
        r[9:12] = np.sqrt(r[9:12])

        u = Utils.hierarchical_solve([jac_r[0:6, :], jac_r[6:12, :]],[-K * r[0:6], -K * r[6:12]])

        q_a += u[0:6] * dt
        q_b += u[6:12] * dt

        robot_a.add_ani_frame(j * dt, q_a)
        robot_b.add_ani_frame(j * dt, q_b)
        plate.add_ani_frame(j * dt, robot_a.fkm() * mth_plate)

        hist_time.append(hist_time[-1] + dt)
        hist_q_a = np.block([hist_q_a, robot_a.q])
        hist_qdot_a = np.block([hist_qdot_a, u[0:6]])
        hist_q_b  = np.block([hist_q_b, robot_b.q])
        hist_qdot_b = np.block([hist_qdot_b, u[6:12]])
        hist_error_rel_pos = np.block([hist_error_rel_pos, r[0:3]])
        error_ori = np.matrix([(180 / (np.pi)) * acos(1 - num * num) for num in r[3:6]]).reshape((3,1))
        hist_error_rel_ori = np.block([hist_error_rel_ori, error_ori])

        j += 1

    # Third part: move down the plate
    mth_a_des = Utils.trn([0, 0, -0.1]) * robot_a.fkm()
    for i in range(round(time_max_1 / dt)):

        if j % 50 == 0 or j == jmax - 1:
            sys.stdout.write('\r')
            sys.stdout.write("[%-20s] %d%%" % ('=' * round(20 * j / (jmax - 1)), round(100 * j / (jmax - 1))))
            sys.stdout.flush()

        r, jac_r = rb.Robot.coop_task_function(robot_a, robot_b, mth_a_des, mth_a_b_des, q_a, q_b)

        r[3:6] = np.sqrt(r[3:6])
        r[9:12] = np.sqrt(r[9:12])

        u = Utils.hierarchical_solve([jac_r[0:6, :], jac_r[6:12, :]],[-K * r[0:6], -K * r[6:12]])

        q_a += u[0:6] * dt
        q_b += u[6:12] * dt

        robot_a.add_ani_frame(j * dt, q_a)
        robot_b.add_ani_frame(j * dt, q_b)
        plate.add_ani_frame(j * dt, robot_a.fkm() * mth_plate)

        hist_time.append(hist_time[-1] + dt)
        hist_q_a = np.block([hist_q_a, robot_a.q])
        hist_qdot_a = np.block([hist_qdot_a, u[0:6]])
        hist_q_b  = np.block([hist_q_b, robot_b.q])
        hist_qdot_b = np.block([hist_qdot_b, u[6:12]])
        hist_error_rel_pos = np.block([hist_error_rel_pos, r[0:3]])
        error_ori = np.matrix([(180 / (np.pi)) * acos(1 - num * num) for num in r[3:6]]).reshape((3,1))
        hist_error_rel_ori = np.block([hist_error_rel_ori, error_ori])

        j += 1

    # Run simulation
    sim.run()

    # Plot graphs
    hist_time.pop()
    Utils.plot(hist_time, hist_q_a, "", "Time (s)", "Joint configuration robot A (rad)", "q")
    Utils.plot(hist_time, hist_qdot_a, "", "Time (s)", "Joint speed robot A (rad/s)", "u")
    Utils.plot(hist_time, hist_q_b, "", "Time (s)", "Joint configuration robot B (rad)", "q")
    Utils.plot(hist_time, hist_qdot_b, "", "Time (s)", "Joint speed robot B (rad/s)", "u")
    Utils.plot(hist_time, hist_error_rel_pos, "", "Time (s)", "Relative position error (m)",
               ['x', 'y', 'z'])
    Utils.plot(hist_time, hist_error_rel_ori, "", "Time (s)", "Relative orientation error (degrees)",
               ['x', 'y', 'z'])

    return sim



from utils import *
import numpy as np


class PointCloud:
    """
   A cloud of points to draw in the simulator.

   There is a fixed set of points ('points' attribute), and animation can be done by choosing, at each time,
   an interval [initial_ind, final_ind] that determines the range of points that will be displayed.

   Parameters
   ----------

   name : string
       The object's name.
       (default: 'genPointCloud').

   size : positive float
       The size of each point in the point cloud.

   color : string
       A HTML-compatible color.

   points : a 3xm numpy array or matrix
       A matrix with 3 rows. The first row is the x coordinates, the second row the y coordinates, and the third row
       the z coordinates.

   """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self):
        """The object name."""
        return self._name

    @property
    def size(self):
        """The size of each point in the point cloud."""
        return self._size

    @property
    def color(self):
        """Color of the object, a HTML-compatible string."""
        return self._color

    @property
    def points(self):
        """The points that compose the point cloud."""
        return np.array(self._points)

    #######################################
    # Constructor
    #######################################

    def __init__(self, name="", points=[], size=0.1, color="blue"):

        # Error handling
        if not Utils.is_a_number(size) or size < 0:
            raise Exception("The parameter 'size' should be a positive float")

        if name=="":
            name="var_pointcloud_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a color")

        if not Utils.is_a_matrix(points, 3):
            raise Exception("The parameter 'points' should be a matrix with 3 rows.")

        # end error handling

        self._name = name
        self._size = size
        self._points = np.matrix(points)
        self._color = color
        self._frames = []
        self._max_time = 0

        self.add_ani_frame(0, 0, np.shape(points)[1])

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Point cloud '" + self._name + "' with " + str(len(self.points)) + " points: \n\n"
        string += " Size: " + str(self.size) + "\n"
        string += " Color: " + str(self._color) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time, initial_ind, final_ind):

        if (not Utils.is_a_number(time)) or time < 0:
            raise Exception("The parameter 'time' should be a nonnegative float.")

        if (not str(type(initial_ind)) == "<class 'int'>") or initial_ind < 0:
            raise Exception("The parameter 'initial_ind' should be a nonnegative integer.")

        if (not str(type(final_ind)) == "<class 'int'>") or final_ind < 0:
            raise Exception("The parameter 'final_ind' should be a nonnegative integer.")

        if initial_ind > final_ind:
            raise Exception("The parameter 'initial_ind' should be greater or equal than the parameter 'final_ind'.")

        if final_ind > len(self.points[0]):
            raise Exception("The parameter 'final_ind' should be at most " + str(len(self.points[0])) + ".")

        # Error handling

        # end error handling

        self._frames.append([time, initial_ind, final_ind])
        self._max_time = max(self._max_time, time)

    def set_ani_frame(self, time, initial_ind, final_ind):

        if (not Utils.is_a_number(time)) or time < 0:
            raise Exception("The parameter 'time' should be a nonnegative float.")

        if (not str(type(initial_ind)) == "<class 'int'>") or initial_ind < 0:
            raise Exception("The parameter 'initial_ind' should be a nonnegative integer.")

        if (not str(type(final_ind)) == "<class 'int'>") or final_ind < 0:
            raise Exception("The parameter 'final_ind' should be a nonnegative integer.")

        if initial_ind > final_ind:
            raise Exception("The parameter 'initial_ind' should be greater or equal than the parameter 'final_ind'.")

        if final_ind > len(self.points[0]):
            raise Exception("The parameter 'final_ind' should be at most " + str(len(self.points[0])) + ".")

        # Error handling

        # end error handling

        self._frames = []
        self.add_ani_frame(0, initial_ind, final_ind)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE POINT CLOUD '" + self.name + "'\n\n"
        string += "const var_" + self.name + " = new PointCloud(" + str(np.around(np.matrix(self.points),4).tolist()) + ", " + str(
            self._frames) + ", '" + self.color + "', " + str(
            self.size) + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string



from utils import *
import numpy as np


# Function used for task function/task Jacobian
def _task_function(self, htm_des, q=None, htm=None):
    if q is None:
        q = self.q

    if htm is None:
        htm = self.htm

    # Error handling
    if not Utils.is_a_matrix(htm_des, 4, 4):
        raise Exception("The parameter 'htm_des' should be a 4x4 homogeneous transformation matrix.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    n = len(self.links)
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    # end error handling

    p_des = htm_des[0:3, 3]
    x_des = htm_des[0:3, 0]
    y_des = htm_des[0:3, 1]
    z_des = htm_des[0:3, 2]

    jac_eef, htm_eef = self.jac_geo(q, "eef", htm)
    p_eef = htm_eef[0:3, 3]
    x_eef = htm_eef[0:3, 0]
    y_eef = htm_eef[0:3, 1]
    z_eef = htm_eef[0:3, 2]

    r = np.matrix(np.zeros((6,1)))
    r[0:3,0] = p_eef - p_des
    r[3] = max(1 - x_des.T * x_eef, 0)
    r[4] = max(1 - y_des.T * y_eef, 0)
    r[5] = max(1 - z_des.T * z_eef, 0)

    n = len(self.links)
    jac_r = np.matrix(np.zeros((6, n)))
    jac_r[0:3, :] = jac_eef[0:3, :]
    jac_r[3, :] = x_des.T * Utils.S(x_eef) * jac_eef[3:6, :]
    jac_r[4, :] = y_des.T * Utils.S(y_eef) * jac_eef[3:6, :]
    jac_r[5, :] = z_des.T * Utils.S(z_eef) * jac_eef[3:6, :]



    return r, jac_r



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_franka_ermika_3(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")


    #For test only
    hpi = 3.14/2
    link_info = [[0, 0, 0, 0, 0, 0, 0],
                 [0.333, 0, 0.316, 0, 0.384, 0, 0],  # "d" translation in z
                 [0, -hpi, hpi, hpi, -hpi, hpi, hpi],  # "alfa" rotation in x
                 [0, 0, 0, 0.0825, -0.0825, 0, 0.088],  # "a" translation in x 0.25
                 [0, 0, 0, 0, 0, 0, 0]]


    H01 = Utils.rotz(link_info[0][0])*Utils.trn([0,0,link_info[1][0]])*Utils.rotx(link_info[2][0])*Utils.trn([link_info[3][0],0,0])
    H02 = H01*Utils.rotz(link_info[0][1]) * Utils.trn([0, 0, link_info[1][1]]) * Utils.rotx(link_info[2][1]) * Utils.trn(
        [link_info[3][1], 0, 0])
    H03 = H02*Utils.rotz(link_info[0][2]) * Utils.trn([0, 0, link_info[1][2]]) * Utils.rotx(link_info[2][2]) * Utils.trn(
        [link_info[3][2], 0, 0])
    H04 = H03*Utils.rotz(link_info[0][3]) * Utils.trn([0, 0, link_info[1][3]]) * Utils.rotx(link_info[2][3]) * Utils.trn(
        [link_info[3][3], 0, 0])
    H05 = H04*Utils.rotz(link_info[0][4]) * Utils.trn([0, 0, link_info[1][4]]) * Utils.rotx(link_info[2][4]) * Utils.trn(
        [link_info[3][4], 0, 0])
    H06 = H05*Utils.rotz(link_info[0][5]) * Utils.trn([0, 0, link_info[1][5]]) * Utils.rotx(link_info[2][5]) * Utils.trn(
        [link_info[3][5], 0, 0])
    H07 = H06*Utils.rotz(link_info[0][6]) * Utils.trn([0, 0, link_info[1][6]]) * Utils.rotx(link_info[2][6]) * Utils.trn(
        [link_info[3][6], 0, 0])
    n = 7
    scale = 1


    # Collision model
    col_model = [[], [], [], [], [], [], []]


    # Create 3d objects


    base_3d_obj = [Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link0.obj',
        scale,
        Utils.trn([0, 0, 0]) ,
        MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],
                     color=color, opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link1.obj',
            scale,
            Utils.trn([0, 0, 0]),
            MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=1))
        ]
    )


    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link2.obj',
            scale,
            Utils.trn([0, 0, 0]),
            MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=1))
        ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/FrankaErmika/link3.obj',
            scale,
            Utils.trn([0, 0, 0]),
            MeshMaterial(metalness=0.3, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=1))
        ]
    )

    link_3d_obj.append(
        [
        ]
    )

    link_3d_obj.append(
        [
        ]
    )
    link_3d_obj.append(
        [
        ]
    )
    link_3d_obj.append(
        [
        ]
    )


    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration

    htm_base_0 = Utils.trn([0, 0, 0])

    q0 = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-170,170],[-120,120],[-170,170],[-120,120],[-170,170],[-120,120],[-175,175]])

    return base_3d_obj, links, np.identity(4), np.identity(4), q0, joint_limits



import numpy as np

class RRTNode:

    #######################################
    # Attributes
    #######################################

    @property
    def q(self):
        return self._q

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        return self._children

    @property
    def value_node(self):
        return self._value_node

    @property
    def index(self):
        return self._index


    #######################################
    # Constructor
    #######################################

    def __init__(self, q, value_node, index):
        self._q = q
        self._parent = None
        self._children = []
        self._value_node = value_node
        self._index = index

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "RRT Node with id " + str(self.index) + ": \n\n"
        string += "Value: "+str(round(self.value_node,5)) + " \n"
        string += "Parent node: "+(str(self.parent.index) if (not self.parent is None) else "root node")+ " \n"
        string += "Children nodes: "+str([node.index for node in self.children])+ " \n"
        string += "Configuration: "+str(self.q.T)

        return string

    #######################################
    # Methods
    #######################################

    def set_parent(self, node_parent):
        if not self._parent is None:
            self._parent._children.remove(self)

        self._parent = node_parent
        node_parent._children.append(self)

    def add_children(self, node_child):
        node_child.set_parent(self)

    def path_to_root(self):
        current_node = self
        list_node=[]
        while not current_node is None:
            list_node.append(current_node)
            current_node = current_node.parent

        list_node.reverse()

        return list_node

class RRTEdge:

    #######################################
    # Attributes
    #######################################

    @property
    def from_node(self):
        return self._from_node

    @property
    def to_node(self):
        return self._to_node

    @property
    def weight(self):
        return self._weight

    #######################################
    # Constructor
    #######################################

    def __init__(self, from_node, to_node, weight=0):
        self._from_node = from_node
        self._to_node = to_node
        self._weight = weight

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "RRT Edge from node " + str(self.from_node.index) + " to "+ str(self.to_node.index) +": \n\n"
        string += "Weight: "+str(round(self.weight,5)) + "\n"
        return string


class RRTTree:

    #######################################
    # Attributes
    #######################################

    @property
    def root_node(self):
        return self._root_node

    @property
    def list_node(self):
        return self._list_node

    @property
    def list_edge(self):
        return self._list_edge

    @property
    def fun_eval_node(self):
        return self._fun_eval_node

    @property
    def fun_dist_node(self):
        return self._fun_dist_node

    @property
    def fun_free(self):
        return self._fun_free

    @property
    def min_dist(self):
        return self._min_dist

    @property
    def step_size(self):
        return self._step_size

    @property
    def q_inf(self):
        return self._q_inf

    @property
    def q_sup(self):
        return self._q_sup

    #######################################
    # Constructor
    #######################################

    def __init__(self, root_q, fun_eval_node, fun_free, q_inf, q_sup,
                 fun_dist_node = lambda q1, q2 : np.linalg.norm(q1-q2), min_dist=0.01, step_size=0.4):

        self._root_node = RRTNode(root_q, fun_eval_node(root_q), 0)
        self._fun_eval_node = fun_eval_node
        self._fun_free = fun_free
        self._fun_dist_node = fun_dist_node
        self._list_node = [self._root_node]
        self._list_edge = []
        self._kd_tree = []
        self._min_dist = min_dist
        self._q_inf = q_inf
        self._q_sup = q_sup
        self._step_size = step_size




    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "RRT Tree with" + str(len(self.list_node)) + " nodes."
        return string

    #######################################
    # Methods
    #######################################

    def nearest_neighbors(self, q_query, k=1):

        list_dist=[]
        for i in range(len(self.list_node)):
            list_dist.append(self.fun_dist_node(q_query, self.list_node[i].q))

        ind = np.argsort(list_dist)

        list_node = []
        for i in range(min(len(list_dist),k)):
            list_node.append(self.list_node[ind[i]])

        return list_node if k > 1 else list_node[-1]

    def sample_free_config(self):


        n = np.shape(self.root_node.q)[0]
        ok = False
        while not ok:
            q_rand = np.matrix(np.zeros( (n,1) ))
            for i in range(n):
                q_rand[i,0] = self.q_inf[i,0] + (self.q_sup[i,0]-self.q_inf[i,0])*np.random.uniform(0,1)

            ok = self.fun_free(q_rand)

        return q_rand

    def config_inside_bounds(self, q):

        inside = True
        n = np.shape(self.root_node.q)[0]
        k=0

        while inside and k < n:
            inside = self.q_sup[k, 0] >= q[k, 0] >= self.q_inf[k, 0]
            k+=1

        return inside

    def add_node_tree(self, k=1):


        if k==1:
            ok = False

            while not ok:

                q_rand = self.sample_free_config()
                node_prox = self.nearest_neighbors(q_rand)
                q_prox = node_prox.q
                dir = (q_rand - q_prox) / (0.0001 + np.linalg.norm(q_rand - q_prox))
                q_target = q_prox + np.random.uniform(0.2 * self.step_size, self.step_size) * dir

                ok = self.config_inside_bounds(q_target) and self.is_path_free(q_prox, q_target)

                if ok:
                    new_node = RRTNode(q_target, self.fun_eval_node(q_target), len(self.list_node))
                    new_node.set_parent(node_prox)
                    new_edge = RRTEdge(node_prox, new_node)
                    self.list_node.append(new_node)
                    self.list_edge.append(new_edge)
        else:
            for i in range(k):
                self.add_node_tree(1)

    def is_path_free(self, q_init, q_end):

        if self.fun_free( (q_init + q_end)/2 ):
            if self.fun_dist_node(q_init, q_end) < self.min_dist:
                return True
            else:
                return self.is_path_free(q_init, (q_init + q_end) / 2) and self.is_path_free((q_init + q_end) / 2,
                                                                                             q_end)
        else:
            return False

    def node_smallest_value(self):

        min_node = self.root_node

        for node in self.list_node:
            if node.value_node < min_node.value_node:
                min_node = node

        return min_node







import numpy as np
from utils import *

_INVHALFPI = 0.63660


def _vector_field(curve, alpha, const_vel):
    # Error handling
    if not Utils.is_a_matrix(curve):
        raise Exception("The parameter 'curve' should be a matrix of float numbers.")

    vector_size =  np.shape(curve)[0]

    if not Utils.is_a_number(alpha) or alpha <= 0:
        raise Exception("The parameter 'alpha' should be a positive float.")

    if not Utils.is_a_number(const_vel):
        raise Exception("The parameter 'const_vel' should be a float.")

    # end error handling

    return lambda p: _vector_field_vel(p, curve, alpha, const_vel, vector_size)


def _vector_field_vel(p, curve, alpha, const_vel, vector_size):
    # Error handling
    if not Utils.is_a_vector(p, vector_size):
        raise Exception("The vector field argument should be a " + str(vector_size) + " dimensional vector.")
    # end error handling

    vec_n, vec_t, min_dist = _compute_ntd(curve, p)
    fun_g = _INVHALFPI * atan(alpha * min_dist)
    fun_h = sqrt(max(1 - fun_g ** 2, 0))
    abs_const_vel = abs(const_vel)
    sgn = const_vel / (abs_const_vel + 0.00001)

    return abs_const_vel * (fun_g * vec_n + sgn * fun_h * vec_t)


def _compute_ntd(curve, p):
    min_dist = float('inf')
    ind_min = -1

    pr = np.matrix(p).reshape((3,1))

    for i in range(np.shape(curve)[1]):
        dist_temp = np.linalg.norm(pr - curve[:,i])
        if dist_temp < min_dist:
            min_dist = dist_temp
            ind_min = i

    vec_n = curve[:,ind_min] - pr
    vec_n = vec_n / (np.linalg.norm(vec_n) + 0.0001)

    if ind_min == np.shape(curve)[1] - 1:
        vec_t = curve[:,1] - curve[:,ind_min]
    else:
        vec_t = curve[:,ind_min + 1] - curve[:,ind_min]

    vec_t = vec_t / (np.linalg.norm(vec_t) + 0.0001)

    return vec_n, vec_t, min_dist



from utils import *
import numpy as np


# (Private function) Used in IK
def _evolve_config(self, q, p_tol, a_tol, htm_target, iter_remain, ignore_orientation):
    n = len(self.links)
    found = False
    zero_u = False
    iter_end = False

    dt = 0.005
    K = 2
    eps = 0.001
    tol_u = 0.00001 / dt
    i = 0

    while (not found) and (not zero_u) and (not iter_end):
        r, jac_r = self.task_function(htm_target, np.array(q))

        if ignore_orientation:
            r = r[0:3]
            jac_r = jac_r[0:3, :]
        else:
            r[3:6] = np.sqrt(r[3:6])

        u = Utils.dp_inv(jac_r, eps) * (-K * r)
        q = q + u * dt

        e_pos = max(abs(r[0:3]))

        if ignore_orientation:
            e_ori = 0
        else:
            e_ori = max([(180 / np.pi) * acos(min(max(1 - num * num, -1), 1)) for num in r[3:6]])

        i += 1

        found = (e_pos < p_tol) and (e_ori < a_tol)
        zero_u = max(abs(u)) < tol_u
        iter_end = i > iter_remain

    return found, i, q


# Inverse kinematics for the end-effector
def _ikm(self, htm_target, q0=None, p_tol=0.005, a_tol=5, no_iter_max=2000, ignore_orientation=False):
    n = len(self._links)
    if q0 is None:
        q0 = np.matrix(np.reshape((2 * np.pi) * np.random.rand(n) - np.pi, (n, 1)))

        # Error handling
    if not Utils.is_a_matrix(htm_target, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not Utils.is_a_natural_number(no_iter_max):
        raise Exception("The parameter 'no_iter_max' should be a nonnegative integer number.")

    if not Utils.is_a_vector(q0, n):
        raise Exception("The parameter 'q0' should be a " + str(n) + " dimensional vector.")

    if (not Utils.is_a_number(p_tol)) or p_tol <= 0:
        raise Exception("The parameter 'p_tol' should be a nonnegative number.")

    if (not Utils.is_a_number(a_tol)) or a_tol <= 0:
        raise Exception("The parameter 'a_tol' should be a nonnegative number.")

    if not str(type(ignore_orientation)) == "<class 'bool'>":
        raise Exception("The parameter 'ignore_orientation' should be a boolean.")

    # end error handling

    j = 0
    found = False
    q = np.matrix(q0).reshape((n, 1))
    no_iter_remain = no_iter_max

    while not found and no_iter_remain >= 0:
        found, i, q = _evolve_config(self, q, p_tol, a_tol, htm_target, no_iter_remain, ignore_orientation)
        no_iter_remain -= i
        if not found:
            q = np.matrix(np.reshape((2 * np.pi) * np.random.rand(n), (n, 1)))

    if not found:
        raise Exception("Solution for IK not found. You can try the following: \n" \
                        " Increasing the maximum number of iterations, 'no_iter_max' (currently " + str(
            no_iter_max) + ")\n" \
                           " Increasing the tolerance for the position, 'p_tol' (currently " + str(p_tol) + " meters)\n" \
                                                                                                            " Increasing the tolerance for the orientation, 'a_tol' (currently " + str(
            a_tol) + " degrees).")
    else:
        return q



from utils import *
import numpy as np


# Function used for task function/task Jacobian
def _coop_task_function(robot_a, robot_b, htm_a_des, htm_a_b_des, q_a, q_b):
    if q_a is None:
        q_a = robot_a.q

    if q_b is None:
        q_b = robot_b.q

    # Error handling
    if not Utils.get_uaibot_type(robot_a) == 'uaibot.Robot':
        raise Exception("The parameter 'robot_a' should be a 'uaibot.Robot' object.")

    if not Utils.is_a_vector(q_a, len(robot_a.links)):
        raise Exception("The parameter 'q_a' should be a " + str(len(robot_a.links)) + " dimensional vector.")

    if not Utils.get_uaibot_type(robot_b) == 'uaibot.Robot':
        raise Exception("The parameter 'robot_b' should be a 'uaibot.Robot' object.")

    if not Utils.is_a_vector(q_a, len(robot_b.links)):
        raise Exception("The parameter 'q_b' should be a " + str(len(robot_b.links)) + " dimensional vector.")

    if not Utils.is_a_matrix(htm_a_des, 4, 4):
        raise Exception("The parameter 'htm_a_des' should be a 4x4 homogeneous transformation matrix.")

    if not Utils.is_a_matrix(htm_a_b_des, 4, 4):
        raise Exception("The parameter 'htm_a_b_des' should be a 4x4 homogeneous transformation matrix.")
    # end error handling

    jac_a, htm_a = robot_a.jac_geo(q_a)
    jac_b, htm_b = robot_b.jac_geo(q_b)
    rot_a = htm_a[0:3, 0:3]
    p_a = htm_a[0:3, 3]
    rot_b = htm_b[0:3, 0:3]
    p_b = htm_b[0:3, 3]

    inv_htm_a = Utils.inv_htm(htm_a)
    htm_a_b = inv_htm_a @ htm_b

    n_a = len(robot_a.links)
    n_b = len(robot_b.links)
    r = np.matrix(np.zeros((12, 1)))
    jac_r = np.matrix(np.zeros((12, n_a + n_b)))

    # First part of the cooperative task function (relative pose between A and B)

    x_rel_des = htm_a_b_des[0:3, 0]
    y_rel_des = htm_a_b_des[0:3, 1]
    z_rel_des = htm_a_b_des[0:3, 2]
    p_rel_des = htm_a_b_des[0:3, 3]

    x_rel = htm_a_b[0:3, 0]
    y_rel = htm_a_b[0:3, 1]
    z_rel = htm_a_b[0:3, 2]
    p_rel = htm_a_b[0:3, 3]

    jac_rel_v = rot_a.T * np.hstack((Utils.S(p_b - p_a) * jac_a[3:6, :] - jac_a[0:3, :], jac_b[0:3, :]))
    jac_rel_w = rot_a.T * np.hstack((-jac_a[3:6, :], jac_b[3:6, :]))

    r[0:3] = p_rel - p_rel_des
    r[3] = max(1 - x_rel_des.T * x_rel, 0)
    r[4] = max(1 - y_rel_des.T * y_rel, 0)
    r[5] = max(1 - z_rel_des.T * z_rel, 0)

    jac_r[0:3, :] = jac_rel_v
    jac_r[3, :] = x_rel_des.T * Utils.S(x_rel) * jac_rel_w
    jac_r[4, :] = y_rel_des.T * Utils.S(y_rel) * jac_rel_w
    jac_r[5, :] = z_rel_des.T * Utils.S(z_rel) * jac_rel_w

    # Second part of the cooperative task function (pose for A)
    x_a_des = htm_a_des[0:3, 0]
    y_a_des = htm_a_des[0:3, 1]
    z_a_des = htm_a_des[0:3, 2]
    p_a_des = htm_a_des[0:3, 3]

    x_a = htm_a[0:3, 0]
    y_a = htm_a[0:3, 1]
    z_a = htm_a[0:3, 2]
    p_a = htm_a[0:3, 3]

    r[6:9] = p_a - p_a_des
    r[9] =  max(1 - x_a_des.T * x_a, 0)
    r[10] = max(1 - y_a_des.T * y_a, 0)
    r[11] = max(1 - z_a_des.T * z_a, 0)

    jac_r[6:9, :] = np.hstack((jac_a[0:3, :], np.zeros((3, n_b))))
    jac_w_a = np.hstack((jac_a[3:6, :], np.zeros((3, n_b))))
    jac_r[9, :] =  x_a_des.T * Utils.S(x_a) * jac_w_a
    jac_r[10, :] = y_a_des.T * Utils.S(y_a) * jac_w_a
    jac_r[11, :] = z_a_des.T * Utils.S(z_a) * jac_w_a

    r = r.reshape((12,1))

    return r, jac_r



import numpy as np


class DistStructLinkObj:

    @property
    def link_number(self):
        """The index of the link in which the collision point is."""
        return self._link_number

    @property
    def link_col_obj_number(self):
        """The index of the collision object of the link in which the collision point is."""
        return self._link_col_obj_number

    @property
    def distance(self):
        """The distance (in meters) between the link and the object."""
        return self._distance

    @property
    def point_link(self):
        """The closest point (witness) in the link. Written in scenario coordinates and in meters."""
        return np.array(self._point_link)

    @property
    def point_object(self):
        """The closest point (witness) in the object. Written in scenario coordinates and in meters."""
        return np.array(self._point_object)

    @property
    def jac_distance(self):
        """The Jacobian of the distance in the robot's configuration space."""
        return np.array(self._jac_distance)

    #######################################
    # Constructor
    #######################################

    def __init__(self, link_number, link_col_obj_number, distance, point_link, point_object, jac_distance):

        self._link_number = link_number
        self._link_col_obj_number = link_col_obj_number
        self._distance = distance
        self._point_link = point_link
        self._point_object = point_object
        self._jac_distance = jac_distance

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string  = "Distance info: \n"
        string += " Link number: "+str(self.link_number)+"\n"
        string += " Collision object of the link number: "+str(self.link_col_obj_number )+"\n"
        string += " Distance: " + str(self.distance) + " m\n"
        string += " Point link: " + str(self.point_link.tolist()) + " m\n"
        string += " Point object: " + str(self.point_object.tolist()) + " m\n"
        string += " Jacobian distance: " + str(self._jac_distance.tolist()) + "\n"
        return string


class DistStructRobotObj:

    #######################################
    # Attributes
    #######################################

    @property
    def obj(self):
        """Return the associated object."""
        return self._obj

    @property
    def robot(self):
        """Return the associated robot."""
        return self._robot

    @property
    def jac_dist_mat(self):
        """
		Return the matrix in which each row we have the distance Jacobian (gradient) for each robot link.
		"""
        return np.matrix(self._jac_dist_mat)

    @property
    def dist_vect(self):
        """
		Return the column vector in which each row we have the distance for each robot link.
		"""
        return np.matrix(self._dist_vect).reshape((self.no_items, 1))

    @property
    def no_items(self):
        """Return the number of items."""
        return self._no_items

    def __getitem__(self, key):
        return self._list_info[key]

    #######################################
    # Constructor
    #######################################

    def __init__(self, obj, robot):

        self._obj = obj
        self._robot = robot
        self._obj_name = obj.name
        self._robot_name = robot.name
        self._no_items = 0

        n = len(robot.links)
        self._list_info = []
        self._jac_dist_mat = np.matrix(np.zeros((0, n)))
        self._dist_vect = []

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        return "Distance struct between robot '" + self.robot.name + "' and object '" \
               + self.obj.name + "', with " + str(self.no_items) + " items"

    #######################################
    # Methods
    #######################################

    def _append(self, link_number, link_col_obj_number, distance, point_link, point_object, jac_distance):

        self._list_info.append(
            DistStructLinkObj(link_number, link_col_obj_number, distance, point_link, point_object, jac_distance))

        self._jac_dist_mat = np.vstack((self._jac_dist_mat, jac_distance))
        self._dist_vect.append(distance)
        self._no_items += 1


    def get_item(self, link_number, link_col_obj_number):
        for dist_info in self._list_info:
            if link_number == dist_info.link_number and link_col_obj_number == dist_info.link_col_obj_number:
                return dist_info

        raise Exception("Item not found!")

    def get_closest_item(self):
        minimum_distance = float('inf')
        index_minimum = -1
        for i in range(self._no_items):
            if self[i].distance < minimum_distance:
                minimum_distance = self[i].distance
                index_minimum = i

        return self[index_minimum]




from setuptools import setup

setup(
   name='uaibot',
   version='0.1',
   description='A useful module',
   author='Vinicius Mariano and Johnata Brayan',
   author_email='foomail@foo.com',
   packages=['foo'],  #same as name
   install_requires=['quadprog', 'colour']
)


from simulation import *
import numpy as np
from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *
from simobjects.pointlight import *
from simobjects.pointcloud import *
from utils import *
from graphics.meshmaterial import *
import robot as rb
import sys


def _control_demo_1():
    # Create simulation and add objects to the scene
    robot = rb.Robot.create_abb_crb(Utils.trn([-0.2,0,0.3]), "robo")
    n = len(robot.links)

    mesh_board = MeshMaterial(roughness=1, metalness=0.9)
    board = Box(htm=Utils.trn([0.6, 0, 0.5]), width=0.05, depth=0.9, height=0.8, color="white",
                mesh_material=mesh_board)
    material_box = MeshMaterial(color="#242526", roughness=1, metalness=1)
    base = Cylinder(htm=Utils.trn([-0.2, 0, 0.15]), radius=0.1, height=0.3, mesh_material = material_box)
    sim = Simulation.create_sim_factory([robot, board, base])


    # Create curve
    theta = np.linspace(0, 2 * np.pi, num=300)
    curve = np.matrix(np.zeros(  (3, len(theta))))
    for i in range(len(theta)):
        t = theta[i]
        curve[:,i] = np.matrix([ [0.56], [0.2 * cos(t)], [0.2 * sin(t) + 0.5]])

    # Create vector field
    vf = rb.Robot.vector_field(curve, 10, 0.3)

    # Parameters
    dt = 0.01
    time_max = 20
    K = 1
    imax = round(time_max / dt)

    # Initializations
    hist_time = []
    hist_qdot = np.matrix(np.zeros((6,0)))
    hist_q = np.matrix(np.zeros((6,0)))
    hist_error_ori = np.matrix(np.zeros((3,0)))

    x_des = np.matrix([0, 0, 1]).reshape((3, 1))
    y_des = np.matrix([0, -1, 0]).reshape((3, 1))
    z_des = np.matrix([1, 0, 0]).reshape((3, 1))

    # Main loop
    draw_points = np.zeros((3, 0))
    reached_board = False

    for i in range(imax):

        if i % 50 == 0 or i == imax - 1:
            sys.stdout.write('\r')
            sys.stdout.write("[%-20s] %d%%" % ('=' * round(20 * i / (imax - 1)), round(100 * i / (imax - 1))))
            sys.stdout.flush()

        jac_eef, htm_eef = robot.jac_geo()

        p_eef = htm_eef[0:3, 3]
        x_eef = htm_eef[0:3, 0]
        y_eef = htm_eef[0:3, 1]
        z_eef = htm_eef[0:3, 2]

        target = np.matrix(np.zeros((6,1)))
        target[0:3] = vf(p_eef)
        target[3] = -K * sqrt(max(1 - x_des.T * x_eef, 0))
        target[4] = -K * sqrt(max(1 - y_des.T * y_eef, 0))
        target[5] = -K * sqrt(max(1 - z_des.T * z_eef, 0))

        jac_target = np.matrix(np.zeros((6, n)))
        jac_target[0:3, :] = jac_eef[0:3, :]
        jac_target[3, :] = x_des.T * Utils.S(x_eef) * jac_eef[3:6, :]
        jac_target[4, :] = y_des.T * Utils.S(y_eef) * jac_eef[3:6, :]
        jac_target[5, :] = z_des.T * Utils.S(z_eef) * jac_eef[3:6, :]

        qdot = Utils.dp_inv(jac_target, 0.002) * target

        q_prox = robot.q + qdot * dt

        robot.add_ani_frame(i * dt, q_prox)

        hist_time.append(i * dt)
        hist_q = np.block([hist_q, robot.q])
        error_ori = np.matrix([(180 / (np.pi)) * acos(1 - min(num * num / (K * K),2)) for num in target[3:6]]).reshape((3,1))
        hist_error_ori = np.block([hist_error_ori, error_ori])
        hist_qdot = np.block([hist_qdot, qdot])

        # See if the end-effector is close to the board to add to the point cloud
        draw_points = np.block([draw_points, p_eef])


        if (not reached_board) and (abs(p_eef[0,0] - board.htm[0,3]) < board.width / 2 + 0.001):
            reached_board = True
            ind_reached = i

    # Set up the cloud of points
    point_cloud = PointCloud(name="drawing", points=draw_points, size=0.025)
    sim.add(point_cloud)
    for i in range(imax):
        if i < ind_reached:
            point_cloud.add_ani_frame(i * dt, 0, 0)
        else:
            point_cloud.add_ani_frame(i * dt, ind_reached, i)

    # Run simulation
    sim.run()

    # Plot graphs
    Utils.plot(hist_time, hist_q, "", "Time (s)", "Joint configuration  (rad)", "q")
    Utils.plot(hist_time, hist_qdot, "", "Time (s)", "Joint speed (rad/s)", "u")
    Utils.plot(hist_time, hist_error_ori, "", "Time (s)", "Orientation error (degrees)", ['x', 'y', 'z'])

    return sim



from utils import *
import numpy as np
from graphics.meshmaterial import *


class Cone:
    """
  A cone object.

  Parameters
  ----------
  htm : 4x4 numpy array or 4x4 nested list
      The object's configuration.
      (default: the same as the current HTM).

  name : string
      The object's name.
      (default: 'genCylinder').

  radius : positive float
      The cylinder base radius, in meters.
      (default: 1).

  height : positive float
      The cylinder's height, in meters.
      (default: 1).

  mass : positive float
      The object's mass, in kg.
      (default: 1).

  color : string
      The object's color, a HTML - compatible string.
      (default: "red").

  opacity : float between 0 and 1
      The opacity. 1 = fully opaque, and 0 = transparent.

  mesh_material: 'MeshMaterial' object
      The object mesh material. If set to 'None', the default is used.
      (default: None).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def radius(self):
        """The cylinder base radius, in meters."""
        return self._radius

    @property
    def height(self):
        """The cylinder height, in meters."""
        return self._height

    @property
    def name(self):
        """The object name."""
        return self._name

    @property
    def htm(self):
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._htm)

    @property
    def mass(self):
        """Mass of the object, in kg."""
        return self._mass

    @property
    def color(self):
        """Color of the object"""
        return self.mesh_material.color

    @property
    def mesh_material(self):
        """Mesh properties of the object"""
        return self._mesh_material

    @property
    def volume(self):
        """The volume of the object, in m³."""
        return self._volume

    #######################################
    # Constructor
    #######################################

    def __init__(self, htm=np.identity(4), name="", radius=1, height=1, mass=1, color="red", opacity=1, \
                 mesh_material=None):

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(mass) or mass < 0:
            raise Exception("The parameter 'mass' should be a positive float.")

        if not Utils.is_a_number(radius) or radius < 0:
            raise Exception("The parameter 'radius' should be a positive float.")

        if not Utils.is_a_number(height) or height < 0:
            raise Exception("The parameter 'height' should be a positive float.")

        if name=="":
            name="var_cylinder_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a HTML-compatible color.")

        if not ((mesh_material is None) or (Utils.get_uaibot_type(mesh_material) == "uaibot.MeshMaterial")):
            raise Exception(
                "The parameter 'mesh_material' should be either 'None' or a 'uaibot.MeshMaterial' object.")

        if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
            raise Exception("The parameter 'opacity' should be a float between 0 and 1.")
        # end error handling

        self._radius = radius
        self._height = height
        self._htm = np.matrix(htm)
        self._name = name
        self._mass = 1
        self._frames = []
        self._volume = np.pi * self.height * self.radius * self.radius
        self._max_time = 0

        if mesh_material is None:
            self._mesh_material = MeshMaterial(color=color, opacity=opacity)
        else:
            self._mesh_material = mesh_material

        # Set initial total configuration
        self.set_ani_frame(self._htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Cone with name '" + self.name + "': \n\n"
        string += " Radius (m): " + str(self.radius) + "\n"
        string += " Height (m): " + str(self.height) + "\n"
        string += " Color: " + str(self.color) + "\n"
        string += " Mass (kg): " + str(self.mass) + "\n"
        string += " HTM: \n" + str(self.htm) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time, htm=None):
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")
        # end error handling

        f = [time, np.around(htm[0,0],4), np.around(htm[0,2],4), np.around(-htm[0,1],4), np.around(htm[0,3],4),
             np.around(htm[1,0],4), np.around(htm[1,2],4), np.around(-htm[1,1],4), np.around(htm[1,3],4),
             np.around(htm[2,0],4), np.around(htm[2,2],4), np.around(-htm[2,1],4), np.around(htm[2,3],4),
             0, 0, 0, 1]

        self._htm = htm
        self._frames.append(f)
        self._max_time = max(self._max_time, time)

    def set_ani_frame(self, htm=None):
        """
    Reset object's animation queue and add a single configuration to the
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        # end error handling

        self._frames = []
        self.add_ani_frame(0, htm)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE CONE '" + self.name + "'\n\n"
        string += self.mesh_material.gen_code(self._name) + "\n"
        string += "const var_" + self._name + " = new Cone(" + str(self._radius) + "," + str(
            self._height) + "," + str(self._frames) + ", material_" + self._name + ");\n"
        string += "sceneElements.push(var_" + self._name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string

    # Compute inertia matrix with respect to the inertia frame
    def inertia_matrix(self, htm=None):
        """
    The 3D inertia matrix of the object, written in the world frame.
    Assume that the transformation between the word frame and the object frame is 'htm'.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration for which the inertia matrix will be computed
        (default: the same as the current HTM).

    Returns
    -------
     inertia_matrix : 3x3 numpy array
        The 3D inertia matrix.
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")
        # end error handling

        Ixx = (1 / 12) * self.mass * (3 * self.radius * self.radius + self.height * self.height)
        Iyy = (1 / 12) * self.mass * (3 * self.radius * self.radius + self.height * self.height)
        Izz = (1 / 2) * self.mass * (self.radius * self.radius)
        Q = htm[0:3, 0:3]
        S = Utils.S(htm[0:3, 3])

        return Q * np.diag([Ixx, Iyy, Izz]) * Q.T - self.mass * S * S

    def copy(self):
        """Return a deep copy of the object, without copying the animation frames."""
        return Cylinder(self.htm, self.name + "_copy", self.radius, self.height, self.mass, self.color)

    def aabb(self):
        """
    Compute the width, depth and height of an axis aligned bounding box (aabb) that
    covers the object. It also considers the current orientation.

    Returns
    -------
     width : positive float
        The width of the box, in meters.

     depth : positive float
        The depth of the box, in meters.

     height : positive float
        The depth of the box, in meters.
    """

        p1 = 2*self.radius * self.htm[:, 0] + 2*self.radius * self.htm[:, 1] + self.height * self.htm[:, 2]
        p2 = -2*self.radius * self.htm[:, 0] + 2*self.radius * self.htm[:, 1] + self.height * self.htm[:, 2]
        p3 = 2*self.radius * self.htm[:, 0] - 2*self.radius * self.htm[:, 1] + self.height * self.htm[:, 2]
        p4 = 2*self.radius * self.htm[:, 0] + 2*self.radius * self.htm[:, 1] - self.height * self.htm[:, 2]

        w = np.max([abs(p1[0, 0]), abs(p2[0, 0]), abs(p3[0, 0]), abs(p4[0, 0])])
        d = np.max([abs(p1[1, 0]), abs(p2[1, 0]), abs(p3[1, 0]), abs(p4[1, 0])])
        h = np.max([abs(p1[2, 0]), abs(p2[2, 0]), abs(p3[2, 0]), abs(p4[2, 0])])

        return w, d, h

    # NEED TO MODIFY
    def generate_samples(self, delta=0.025):

        P = np.matrix(np.zeros((3, 0)))

        T = round(2*np.pi*self.radius / delta)+1
        R = round(self.radius/delta)+1
        H = round(self.height / delta)+1


        for i in range(T):
            u = (2*np.pi)*i/(T-1)
            for j in range(H):
                v = j/(H-1)

                x = self.radius*np.cos(u)
                y = self.radius*np.sin(u)
                z = (-self.height/2 + v*self.height)
                P = np.block([P, np.matrix([x,y,z]).transpose()])


        for i in range(R):
            v = self.radius * (i/(R-1))
            T = round(2 * np.pi * v / delta)
            for j in range(T):
                u = (2*np.pi)*j/(T-1)

                x = v * np.cos(u)
                y = v * np.sin(u)
                z = -self.height / 2
                P = np.block([P, np.matrix([x, y, z]).transpose()])

                x = v * np.cos(u)
                y = v * np.sin(u)
                z = self.height / 2
                P = np.block([P, np.matrix([x, y, z]).transpose()])

        for i in range(np.shape(P)[1]):
            P[:,i] = self.htm[0:3,0:3]*P[:,i]+self.htm[0:3,-1]

        return P

    # Compute the projection of a point into an object
    #NEED TO MODIFY
    def projection(self, point, htm=None):
        """
    The projection of a point in the object, that is, the
    closest point in the object to a point 'point'.

    Parameters
    ----------
    point : 3D vector
        The point for which the projection will be computed.

    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
     proj_point : 3D vector
        The projection of the point 'point' in the object.

     d : positive float
        The distance between the object and 'point'.
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix")

        if not Utils.is_a_vector(point, 3):
            raise Exception("The parameter 'point' should be a 3D vector")

        # end error handling
        tpoint = htm[0:3, 0:3].T * (point - htm[0:3, 3])


        r = sqrt(tpoint[0,0] ** 2 + tpoint[1,0] ** 2)

        if r < self.radius:
            x = tpoint[0,0]
            y = tpoint[1,0]
            dr2=0
        else:
            x = self.radius * tpoint[0,0] / r
            y = self.radius * tpoint[1,0] / r
            dr2 = (r-self.radius)**2

        if abs(tpoint[2,0]) < self.height/2:
            z = tpoint[2,0]
            dz2 = 0
        else:
            z = self.height/2 if tpoint[2,0] > 0 else -self.height/2
            dz2 = (abs(tpoint[2,0]) - self.height/2)**2

        d = sqrt(dr2+dz2)

        return htm[0:3, 0:3] * np.matrix([[x], [y], [z]]) + htm[0:3, 3], d



import gym
from gym import spaces
from simulation import *
import numpy as np
from utils import *



class RobotEnv(gym.Env):
    """Robot Environment that follows gym interface"""

    metadata = {'render.modes': ['human']}

    def __init__(self, robot, obstacle, mthd, minqdot=-5, maxqdot=5):

        n = len(robot.links)
        self.robot = robot
        self.mthd = mthd
        self.obstacle = obstacle
        self.oldDistStruct = None
        self.h = 0.001
        self.current_step = 0
        self.last_envInfo = np.zeros((12,))
        self.last_action = np.zeros((n,))
        self.homeq = robot.q

        self.reward_range = (-30, 30)
        # Create action space (joint speed)
        self.action_space = spaces.Box(
            low=minqdot * np.ones((n,)), high=maxqdot * np.ones((n,)), dtype=np.float16)

        self.observation_space = spaces.Box(
            low=-10, high=10, shape=(24 + n,), dtype=np.float16)

    def next_observation(self):
        n = len(self.robot.links)
        r = self.robot.task_function(self.mthd)[0]
        ds = self.robot.compute_dist(obj=self.obstacle, h=self.h, g=self.h, old_dist_struct=self.oldDistStruct)
        obs = np.zeros((24 + n,))
        envInfo = np.zeros((12,))

        envInfo[0:6] = r
        envInfo[6:9] = ds.get_closest_item()["pObjCol"].reshape((3,))
        envInfo[9:12] = ds.get_closest_item()["pObj"].reshape((3,))

        obs[0:12] = self.last_envInfo
        obs[12:24] = envInfo
        obs[24:24 + n] = self.last_action

        self.last_envInfo = envInfo

        return obs

    def reward(obs):
        kTaskPos = 10
        kTaskOri = 10
        kObs = 3

        rTaskPos = kTaskPos * (np.linalg.norm(obs[0:3]) - np.linalg.norm(obs[12:15]))
        rTaskOri = kTaskOri * (np.linalg.norm(obs[3:6]) - np.linalg.norm(obs[15:18]))
        # rObs = -kObs*(np.linalg.norm(obs[6:9])-np.linalg.norm(obs[15:18]))

        if max(abs(obs[12:18])) <= 0.01:
            print("Task achieved")
            return 10
        else:
            if np.linalg.norm(obs[18:21] - obs[21:24]) <= 0.005:
                print("Collision")
                return -10
            else:
                return rTaskPos + rTaskOri

    def isDone(obs):
        return np.linalg.norm(obs[18:21] - obs[21:24]) <= 0.005 or max(abs(obs[12:18])) <= 0.01

    def step(self, action):
        # Execute one time step within the environment
        self.current_step += 1
        dt = 0.01
        self.robot.set_ani_frame(self.robot.q + action * dt)
        self.last_action = action
        obs = self.next_observation()

        return obs, RobotEnv.reward(obs), RobotEnv.isDone(obs), {}

    def reset(self):
        n = len(self.robot.links)

        found_config = False

        while not found_config:
            q = np.random.uniform(0,6.28,(6,))
            dist = self.robot.compute_dist(q = q, obj=self.obstacle).get_closest_item()["hgDistance"]
            found_config = (dist>0.02)


        self.robot.set_ani_frame(q)
        self.current_step = 0
        self.last_task = np.zeros((9,))
        self.last_action = np.zeros((n,))

        obs = self.next_observation()
        print(np.linalg.norm(obs[18:21] - obs[21:24]))
        return obs

    def render(self, mode='human', close=False):
        # Render the environment to the screen
        print("Current task status:")
        print(self.last_task)
        print("Current configuration:")
        print(self.robot.q)



from utils import *
import numpy as np
from ._dist_struct_robot_auto import DistStructRobotAuto


# Compute the distance from each non-sequential link to other link
def _compute_dist_auto(self, q=None, old_dist_struct=None, tol=0.0005, no_iter_max=20, max_dist = np.inf):
    n = len(self.links)

    if q is None:
        q = self.q

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_number(tol) or tol <= 0:
        raise Exception("The parameter 'tol' should be a positive number.")

    if not Utils.is_a_number(max_dist) or tol <= 0:
        raise Exception("The parameter 'max_dist' should be a positive number, or np.inf.")

    if not Utils.is_a_natural_number(no_iter_max) or no_iter_max <= 0:
        raise Exception("The parameter 'no_iter_max' should be a positive natural number.")

    if not (old_dist_struct is None):
        try:
            if not id(old_dist_struct.robot) == id(self):
                Exception("The parameter 'old_dist_struct' is a 'DistStructRobotAuto' object, but it " \
                          "must have to be relative to the SAME robot object, and " \
                          "this is not the case.")
        except:
            raise Exception("The parameter 'old_dist_struct' must be a 'DistStructRobotAuto' object.")

    # end error handling

    dist_struct = DistStructRobotAuto(self)

    jac_dh, mth_dh = self.jac_geo(q, "dh")

    col_object_copy = []

    # Update all collision objects of all links
    for i in range(n):
        col_object_copy.append([])
        for j in range(len(self.links[i].col_objects)):
            temp_copy = self.links[i].col_objects[j][0].copy()
            htmd = self.links[i].col_objects[j][1]
            temp_copy.set_ani_frame(mth_dh[i][:, :] * htmd)
            col_object_copy[i].append(temp_copy)

    # Compute the distance structure
    for i in range(n):
        for j in range(i+2,n):
            for isub in range(len(self.links[i].col_objects)):
                for jsub in range(len(self.links[j].col_objects)):

                    est_dist = 0 if math.isinf(max_dist) else Utils.compute_aabbdist(col_object_copy[i][isub], col_object_copy[j][jsub])

                    if est_dist <= max_dist:

                        if old_dist_struct is None:
                            p_obj_0 = np.matrix(np.random.uniform(-100, 100, size=(3, 1)))
                        else:
                            p_obj_0 = old_dist_struct.get_item(i, isub, j, jsub).point_object

                        p_obj_i, p_obj_j, d = Utils.compute_dist(col_object_copy[i][isub], col_object_copy[j][jsub] \
                                                                 , p_obj_0, tol, no_iter_max)

                        jac_obj_i = jac_dh[i][0:3, :] - Utils.S(p_obj_i - mth_dh[i][0:3, 3]) * jac_dh[i][3:6, :]
                        jac_obj_j = jac_dh[j][0:3, :] - Utils.S(p_obj_j - mth_dh[j][0:3, 3]) * jac_dh[j][3:6, :]

                        jac_dist = (p_obj_i - p_obj_j).T * (jac_obj_i - jac_obj_j) / d
                        dist_struct._append(i, isub, j, jsub, d, p_obj_i, p_obj_j, jac_dist)



    return dist_struct



from simulation import *
import numpy as np
from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *
from simobjects.pointlight import *
from simobjects.pointcloud import *
from simobjects.frame import *
from utils import *
import robot as rb
from cvxopt import matrix, solvers
from graphics.meshmaterial import *
import sys
import time


def _control_demo_3():
    solvers.options['show_progress'] = False
    np.set_printoptions(precision=4, suppress=True, linewidth=150)

    robot = rb.Robot.create_kuka_lbr_iiwa()

    texture_wall = Texture(
        url='https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_6.stl',
        wrap_s='RepeatWrapping', wrap_t='RepeatWrapping', repeat=[4, 4])

    material_wall = MeshMaterial(texture_map=texture_wall, roughness=1, metalness=1)

    wall1 = Box(name="wall1", htm=Utils.trn([0.3, -0.5, 0.7]), width=0.05, depth=0.6, height=1.4,
                mesh_material=material_wall)
    wall2 = Box(name="wall2", htm=Utils.trn([0.3, 0.5, 0.7]), width=0.05, depth=0.6, height=1.4,
                mesh_material=material_wall)
    wall3 = Box(name="wall3", htm=Utils.trn([0.3, 0, 0.25]), width=0.05, depth=0.4, height=0.5,
                mesh_material=material_wall)
    wall4 = Box(name="wall4", htm=Utils.trn([0.3, 0, 1.15]), width=0.05, depth=0.4, height=0.5,
                mesh_material=material_wall)

    def evaluate_error(r):
        error_pos = max(abs(r[0:3]))
        error_ori = (180 / np.pi) * max(abs(np.arccos(1 - r[3:6])))
        ok1 = error_pos < 0.005
        ok2 = error_ori < 5 if len(r.tolist()) > 3 else True

        return ok1 and ok2, error_pos, error_ori

    def dist_computation(q, old_struct):
        dist_wall_1 = robot.compute_dist(obj=wall1, q=q, old_dist_struct=old_struct[0])
        dist_wall_2 = robot.compute_dist(obj=wall2, q=q, old_dist_struct=old_struct[1])
        dist_wall_3 = robot.compute_dist(obj=wall3, q=q, old_dist_struct=old_struct[2])
        dist_wall_4 = robot.compute_dist(obj=wall4, q=q, old_dist_struct=old_struct[3])

        struct = [dist_wall_1, dist_wall_2, dist_wall_3, dist_wall_4]

        jac_dist = np.block(
            [[dist_wall_1.jac_dist_mat], [dist_wall_2.jac_dist_mat], [dist_wall_3.jac_dist_mat],
             [dist_wall_4.jac_dist_mat]])
        dist_vect = np.block(
            [[dist_wall_1.dist_vect], [dist_wall_2.dist_vect], [dist_wall_3.dist_vect], [dist_wall_4.dist_vect]])

        return dist_vect, jac_dist, struct

    # Target pose definition
    pose_tg = []
    pose_tg.append(Utils.trn([0.5, -0.3, 0.7]) @ Utils.rotx(3.14 / 2))
    pose_tg.append(Utils.trn([0.5, 0, 0.7]) @ Utils.roty(3.14 / 2))
    pose_tg.append(robot.fkm(axis="eef"))

    # Create simulation
    sim = Simulation.create_sim_factory([robot, wall1, wall2, wall3, wall4])

    for k in range(len(pose_tg)):
        sim.add(Frame(name="pose_tg_" + str(k), htm=pose_tg[k]))

    # Parameters
    dt = 0.03
    alpha = 2
    beta = 0.005
    gamma = 0.5
    sigma = 0.5
    dist_safe = 0.2
    qdot_max = (np.pi / 180) * np.matrix([[85], [85], [100], [75], [130], [135], [135]])
    qdot_min = -qdot_max
    xi = 0.5
    h = 0.05
    q = robot.q
    qdot = np.zeros((7, 1))

    # Initializations
    struct = [None, None, None, None]

    i = 0
    imax = round(47.4/dt)

    hist_dist = []
    hist_time = []
    hist_r = np.matrix(np.zeros((6,0)))
    hist_q = np.matrix(np.zeros((7, 0)))
    hist_qddot = np.matrix(np.zeros((7,0)))
    hist_qdot = np.matrix(np.zeros((7,0)))
    hist_V = []

    error_qp = False
    iteration_end = False

    #Main loop
    for k in range(len(pose_tg)):
        converged = False
        while not converged and not error_qp and not iteration_end:

            #This is just for showing progress when the simulation is run
            if i % 50 == 0 or i == imax - 1:
                sys.stdout.write('\r')
                sys.stdout.write("[%-20s] %d%%" % ('=' * round(20 * i / (imax - 1)), round(100 * i / (imax - 1))))
                sys.stdout.flush()

            #Compute r and r_dot
            [r, jac_r] = robot.task_function(pose_tg[k], q=q)
            [r_next, jac_r_next] = robot.task_function(pose_tg[k], q=q + qdot * dt)

            r_dot = (r_next - r) / dt
            jac_r_dot = (jac_r_next - jac_r) / dt

            #Compute dist_vect, dist_vect_dot and the distance Jacobian

            dist_vect, jac_dist, struct = dist_computation(q, struct)
            dist_vect_next, jac_dist_next, struct = dist_computation(q + qdot * dt, struct)

            dist_vect_dot = (dist_vect_next - dist_vect) / dt
            jac_dist_dot = (jac_dist_next - jac_dist) / dt

            #Create the quadratic program parameters
            A = jac_dist
            b = -jac_dist_dot @ qdot -2 * gamma * dist_vect_dot - (gamma ** 2) * (dist_vect - dist_safe)
            H = 2 * (np.transpose(jac_r) * jac_r) + 2 * beta * np.identity(7)
            f = 2 * np.transpose(jac_r) * (
                    jac_r_dot @ qdot + 2 * alpha * r_dot + (alpha ** 2) * r) + 2 * beta * sigma * qdot

            A = np.block([[A], [np.identity(7)]])
            A = np.block([[A], [-np.identity(7)]])


            b = np.block([[b], [-xi * (qdot-qdot_min)], [xi * (qdot-qdot_max)]  ])


            #Solve the quadratic program
            try:
                qddot = solvers.qp(matrix(H), matrix(f), matrix(-A), matrix(-b))['x']
            except:
                qddot = np.matrix(np.zeros((7, 1)))
                error_qp = True

            qddot = np.reshape(qddot, (7, 1))

            #First order explicit Euler simulation
            q += qdot * dt
            qdot += qddot * dt

            #Add animation to simulation
            robot.add_ani_frame(i * dt, q)

            #Store data for showing the graphs later
            dist_vect, _, _ = dist_computation(q, struct)

            hist_time.append(i * dt)
            hist_dist.append(np.amin(dist_vect))
            hist_r = np.block([hist_r,r])
            hist_q = np.block([hist_q,q])
            hist_qdot = np.block([hist_qdot,qdot])
            hist_qddot = np.block([hist_qddot,qddot])
            V=alpha * (np.linalg.norm(r_dot + alpha * r) ** 2) + sigma * beta * (np.linalg.norm(qdot) ** 2)
            hist_V.append(V)

            #Continue the loop, check if converged
            i += 1
            converged, error_pos, error_ori = evaluate_error(r)

            #iteration_end  = i>500

    # Run simulation
    sim.run()

    # Plot graphs
    Utils.plot(hist_time, hist_dist, "", "Time (s)", "True distance (m)", "dist")
    Utils.plot(hist_time, hist_q, "", "Time (s)", "Joint configuration (rad)", "q")
    Utils.plot(hist_time, hist_qdot, "", "Time (s)", "Joint speed (rad/s)", "qdot")
    Utils.plot(hist_time, hist_qddot, "", "Time (s)", "Joint acceleration (rad/s²)", "u")
    Utils.plot(hist_time, hist_V, "", "Time (s)", "Lyapunov function", "V")
    fig = Utils.plot(hist_time, hist_r, "", "Time (s)", "Task function",
                     ["posx", "posy", "posz", "orix", "oriy", "oriz"])

    return sim